import { writeFileSync } from 'fs';
import { Readable } from 'stream';
import { promisify } from 'util';

import axios from 'axios';
import * as CallableInstance from 'callable-instance';
import { XmlEntities } from 'html-entities';
import { PNG } from 'pngjs';
import * as puppeteer from 'playwright';
import * as sharp from 'sharp';
import * as temp from 'temp';

import { getLogger } from './loggers';
import { Message } from './koishi';
import { MediaEntity, Tweets } from './twitter';
import { chainPromises } from './utils';

const xmlEntities = new XmlEntities();

const ZHType = (type: string) => new class extends String {
  public type = super.toString();
  public toString = () => `[${super.toString()}]`;
}(type);

const typeInZH = {
  photo: ZHType('图片'),
  video: ZHType('视频'),
  animated_gif: ZHType('GIF'),
};

const logger = getLogger('webshot');

class Webshot extends CallableInstance<[Tweets, (...args) => void, number], Promise<void>> {

  private browser: puppeteer.Browser;
  private mode: number;
  private wsUrl: string;

  constructor(wsUrl: string, mode: number, onready?: (...args) => void) {
    super('webshot');
    // tslint:disable-next-line: no-conditional-assignment
    // eslint-disable-next-line no-cond-assign
    if (this.mode = mode) {
      onready();
    } else {
      this.wsUrl = wsUrl;
      this.connect(onready);
    }
  }

  private connect = (onready?: (...args) => void): Promise<void> =>
    axios.get<{[key in 'chromium' | 'firefox' | 'webkit']?: string}>(this.wsUrl)
      .then(res => {
        logger.info(`received websocket endpoint: ${JSON.stringify(res.data)}`);
        const browserType = Object.keys(res.data)[0] as keyof typeof res.data;
        return (puppeteer[browserType] as puppeteer.BrowserType<puppeteer.Browser>)
          .connect({wsEndpoint: res.data[browserType]});
      })
      .then(browser => this.browser = browser)
      .then(() => {
        logger.info('launched puppeteer browser');
        if (onready) return onready();
      })
      .catch(error => this.reconnect(error, onready));

  private reconnect = (error, onready?: (...args) => void) => {
    logger.error(`connection error, reason: ${error}`);
    logger.warn('trying to reconnect in 2.5s...');
    return promisify(setTimeout)(2500)
      .then(() => this.connect(onready));
  };

  private extendEntity = (media: MediaEntity) => {
    logger.info('not working on a tweet');
  };

  private renderWebshot = (
    url: string, height: number, webshotDelay: number,
    ...morePostProcessings: ((page: puppeteer.Page) => Promise<any>)[]
  ): Promise<string> => {
    temp.track();
    const jpeg = (data: Readable) => data.pipe(sharp()).jpeg({quality: 90, trellisQuantisation: true});
    const sharpToFile = (pic: sharp.Sharp) => new Promise<string>(resolve => {
      const webshotTempFilePath = temp.path({suffix: '.jpg'});
      pic.toFile(webshotTempFilePath).then(() => resolve(`file://${webshotTempFilePath}`));
    });
    const promise = new Promise<{ path: string, boundary: null | number }>((resolve, reject) => {
      const width = 720;
      const zoomFactor = 2;
      logger.info(`shooting ${width}*${height} webshot for ${url}`);
      this.browser.newPage({
        bypassCSP: true,
        deviceScaleFactor: zoomFactor,
        locale: 'ja-JP',
        timezoneId: 'Asia/Tokyo',
        userAgent: 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/67.0.3396.99 Safari/537.36',
      })
        .then(page => {
          const startTime = new Date().getTime();
          const getTimerTime = () => new Date().getTime() - startTime;
          const getTimeout = () => Math.max(500, webshotDelay - getTimerTime());
          page.setViewportSize({
            width: width / zoomFactor,
            height: height / zoomFactor,
          })
            .then(() => page.goto(url, {waitUntil: 'load', timeout: getTimeout()}))
            // hide header, "more options" button, like and retweet count
            .then(() => page.addStyleTag({
              content: 'header{display:none!important}[data-testid="caret"],[data-testid="tweet"]+*>:nth-last-child(-n+2){display:none}',
            }))
            .then(() => page.addStyleTag({
              content: '*{font-family:-apple-system,".Helvetica Neue DeskInterface",Hiragino Sans,Hiragino Sans GB,sans-serif!important}',
            }))
            .then(() => page.addStyleTag({
              content: '*{font-family:-apple-system,".Helvetica Neue DeskInterface",Hiragino Sans,Hiragino Sans GB,sans-serif!important}',
            }))
            // remove listeners
            .then(() => page.evaluate(() => {
              const poll = setInterval(() => {
                document.querySelectorAll('div[data-testid="placementTracking"]').forEach(container => {
                  if (container) {
                    container.innerHTML = container.innerHTML;
                    clearInterval(poll);
                  }
                });
              }, 250);
            }))
            .then(() => page.waitForSelector('article', {timeout: getTimeout()}))
            .catch((err: Error): Promise<puppeteer.ElementHandle<Element> | null> => {
              if (err.name !== 'TimeoutError') throw err;
              logger.warn(`navigation timed out at ${getTimerTime()} seconds`);
              return null;
            })
            .then(handle => {
              if (handle === null) throw new puppeteer.errors.TimeoutError();
            })
            .then(() => page.evaluate(() => {
              const cardImg = document.querySelector('div[data-testid^="card.layout"][data-testid$=".media"] img');
              if (typeof cardImg?.getAttribute('src') === 'string') {
                const match = /^(.*\/card_img\/(\d+)\/.+\?format=.*)&name=/.exec(cardImg?.getAttribute('src'));
                if (match) {
                  // tslint:disable-next-line: variable-name
                  const [media_url_https, id_str] = match.slice(1);
                  return {
                    media_url: media_url_https.replace(/^https/, 'http'),
                    media_url_https,
                    url: '',
                    display_url: '',
                    expanded_url: '',
                    type: 'photo',
                    id: Number(id_str),
                    id_str,
                    sizes: undefined,
                  };
                }
              }
            }))
            .then(cardImg => {
              if (cardImg) this.extendEntity(cardImg); 
            })
            .then(() => page.addScriptTag({
              content: 'document.documentElement.scrollTop=0;',
            }))
            .then(() => chainPromises(morePostProcessings.map(func => () => func(page))))
            .then(() => promisify(setTimeout)(getTimeout()))
            .then(() => page.screenshot())
            .then(screenshot => {
              new PNG({
                filterType: 4,
                deflateLevel: 0,
              }).on('parsed', function () {
                // remove comment area
                // tslint:disable-next-line: no-shadowed-variable
                // eslint-disable-next-line @typescript-eslint/no-shadow
                const idx = (x: number, y: number) => (this.width * y + x) << 2;
                let boundary: number = null;
                const x = zoomFactor * 2;
                for (let y = 0; y < this.height; y += zoomFactor) {
                  if (
                    this.data[idx(x, y)] !== 255 &&
                    this.data[idx(x, y)] === this.data[idx(x + zoomFactor * 10, y)]
                  ) {
                    if (this.data[idx(x, y + 18 * zoomFactor)] !== 255) {
                      // footer kicks in
                      boundary = null;
                    } else {
                      boundary = y;
                    }
                    break;
                  }
                }
                if (boundary !== null) {
                  logger.info(`found boundary at ${boundary}, cropping image`);
                  this.data = this.data.slice(0, idx(this.width, boundary));
                  this.height = boundary;

                  sharpToFile(jpeg(this.pack())).then(path => {
                    logger.info(`finished webshot for ${url}`);
                    resolve({path, boundary});
                  });
                } else if (height >= 8 * 1920) {
                  logger.warn('too large, consider as a bug, returning');
                  sharpToFile(jpeg(this.pack())).then(path => {
                    resolve({path, boundary: 0});
                  });
                } else {
                  logger.info('unable to find boundary, try shooting a larger image');
                  resolve({path: '', boundary});
                }
              }).parse(screenshot);
            })
            .catch(err => {
              if (err instanceof Error && err.name !== 'TimeoutError') throw err;
              logger.error(`error shooting webshot for ${url}, could not load web page of tweet`);
              resolve({path: '', boundary: 0});
            })
            .finally(() => { page.close(); });
        })
        .catch(reject);
    });
    return promise.then(data => {
      if (data.boundary === null) {
        return this.renderWebshot(url, height + 1920, webshotDelay, ...morePostProcessings);
      } else return data.path;
    }).catch(error => this.reconnect(error)
      .then(() => this.renderWebshot(url, height, webshotDelay, ...morePostProcessings))
    );
  };

  private fetchMedia = (url: string): Promise<string> => new Promise<ArrayBuffer>((resolve, reject) => {
    logger.info(`fetching ${url}`);
    axios({
      method: 'get',
      url,
      responseType: 'arraybuffer',
      timeout: 150000,
    }).then(res => {
      if (res.status === 200) {
        logger.info(`successfully fetched ${url}`);
        resolve(res.data);
      } else {
        logger.error(`failed to fetch ${url}: ${res.status}`);
        reject();
      }
    }).catch (err => {
      logger.error(`failed to fetch ${url}: ${err instanceof Error ? err.message : err}`);
      reject();
    });
  }).then(data =>
    (ext => {
      const mediaTempFilePath = temp.path({suffix: `.${ext}`});
      writeFileSync(mediaTempFilePath, Buffer.from(data));
      const path = `file://${mediaTempFilePath}`;
      switch (ext) {
        case 'jpg':
        case 'png':
          return Message.Image(path);
        case 'mp4':
          return Message.Video(path);
      }
      logger.warn('unable to find MIME type of fetched media, failing this fetch');
      throw Error();
    })(((/\?format=([a-z]+)&/.exec(url)) ?? (/.*\/.*\.([^?]+)/.exec(url)))[1])
  );

  public webshot(
    tweets: Tweets,
    callback: (msgs: string, text: string, author: string) => void,
    webshotDelay: number
  ): Promise<void> {
    let promise = new Promise<void>(resolve => {
      resolve();
    });
    tweets.forEach(twi => {
      promise = promise.then(() => {
        logger.info(`working on ${twi.user.screen_name}/${twi.id_str}`);
      });
      const originTwi = twi.retweeted_status || twi;
      let messageChain = '';

      // text processing
      let author = `${twi.user.name} (@${twi.user.screen_name}):\n`;
      if (twi.retweeted_status) author += `RT @${twi.retweeted_status.user.screen_name}: `;

      let text = originTwi.full_text;

      promise = promise.then(() => {
        if (originTwi.entities && originTwi.entities.urls && originTwi.entities.urls.length) {
          originTwi.entities.urls.forEach(url => {
            text = text.replace(new RegExp(url.url, 'gm'), url.expanded_url);
          });
        }
        if (originTwi.extended_entities) {
          originTwi.extended_entities.media.forEach(media => {
            text = text.replace(new RegExp(media.url, 'gm'), this.mode === 1 ? typeInZH[media.type] : '');
          });
        }
        if (this.mode > 0) messageChain += (author + xmlEntities.decode(text));
      });

      // invoke webshot
      if (this.mode === 0) {
        const url = `https://mobile.twitter.com/${twi.user.screen_name}/status/${twi.id_str}`;
        this.extendEntity = (cardImg: MediaEntity) => {
          originTwi.extended_entities = {
            ...originTwi.extended_entities,
            media: [
              ...originTwi.extended_entities?.media ?? [],
              cardImg,
            ],
          };
        };
        promise = promise.then(() => this.renderWebshot(url, 1920, webshotDelay))
          .then(fileurl => {
            if (fileurl) return Message.Image(fileurl);
            return author + text;
          })
          .then(msg => {
            if (msg) messageChain += msg;
          });
      }
      // fetch extra entities
      // tslint:disable-next-line: curly
      // eslint-disable-next-line curly
      if (1 - this.mode % 2) promise = promise.then(() => {
        if (originTwi.extended_entities) {
          return chainPromises(originTwi.extended_entities.media.map(media => () => {
            let url: string;
            if (media.type === 'photo') {
              url = media.media_url_https.replace(/\.([a-z]+)$/, '?format=$1') + '&name=orig';
            } else {
              url = media.video_info.variants
                .filter(variant => variant.bitrate !== undefined)
                .sort((var1, var2) => var2.bitrate - var1.bitrate)
                .map(variant => variant.url)[0]; // largest video
            }
            const altMessage = `\n[失败的${typeInZH[media.type as keyof typeof typeInZH].type}：${url}]`;
            return this.fetchMedia(url)
              .catch(error => {
                logger.warn('unable to fetch media, sending plain text instead...');
                return altMessage;
              })
              .then(msg => { messageChain += msg; });
          }));
        }
      });
      // append URLs, if any
      if (this.mode === 0) {
        if (originTwi.entities && originTwi.entities.urls && originTwi.entities.urls.length) {
          promise = promise.then(() => {
            const urls = originTwi.entities.urls
              .filter(urlObj => urlObj.indices[0] < originTwi.display_text_range[1])
              .map(urlObj => `\n\ud83d\udd17 ${urlObj.expanded_url}`);
            if (urls.length) {
              messageChain += urls.join('');
            }
          });
        }
      }
      // refer to quoted tweet, if any
      if (originTwi.is_quote_status) {
        promise = promise.then(() => {
          const match = /\/status\/(\d+)/.exec(originTwi.quoted_status_permalink?.expanded);
          const blockQuoteIdStr = match ? match[1] : originTwi.quoted_status?.id_str;
          if (blockQuoteIdStr) messageChain += `\n回复此命令查看引用的推文：\n/twitter_view ${blockQuoteIdStr}`;
        });
      }
      promise.then(() => {
        logger.info(`done working on ${twi.user.screen_name}/${twi.id_str}, message chain:`);
        logger.info(JSON.stringify(Message.ellipseBase64(messageChain)));
        callback(messageChain, xmlEntities.decode(text), author);
      });
    });
    return promise;
  }
}

export default Webshot;
